'use strict';

const axios = require("axios"),
    authUtils = require("http-auth-utils"),
    Q = require("q");

module.exports = {
    /**
     * Starts to read the stream the "success" callback is called on every frame
     * @param success(base64)
     * @param error(error)
     * @param url
     * @param user
     * @param password
     */
    startStream(success, error, url, user, password) {
        let probeAbortController = new AbortController();

        this._successCallback = b64String => {
            if (this._readyForFrame) {
                this._readyForFrame = false;
                success(b64String);
            }
        };
        this._errorCallback = err => {
            this.stopStream();
            error(err);
        };

        this._getAuthHeader({
            url: url,
            user: user,
            pwd: password
        }).then(authValue => {
            let config = {
                responseType: "stream",
                signal: probeAbortController.signal
            };
            if (authValue) {
                config.headers = config.headers || {};
                config.headers["Authorization"] = authValue;
            }
            // Validate the contentType to prevent us from downloading a MJPEG stream instead of JPG
            axios.get(url, config).then((resp) => {
                probeAbortController.abort();
                probeAbortController.signal.onabort = () => {
                    resp.data.req.abort();
                    resp.data.req.destroy();
                };
                let contentType = resp.headers["content-type"];
                this.stopStream();
                this._abortHandler = new AbortController();
                if (contentType.startsWith("multipart/x-mixed-replace")) {
                    this._fetchMjpgFrame({
                        url: url,
                        headers: config.headers
                    });
                } else if (contentType === "image/jpeg") {
                    this._fetchJpg({
                        url: url,
                        headers: config.headers
                    });
                } else {
                    this._errorCallback(`Unsupported contentType: ${contentType}`);
                }
            }, err => {
                if (typeof err === "object" && err.message === "Canceled") {
                    return; // Ignore, its due to the probeAbortController
                }
                this._errorCallback(err.message)
            });
        }, this._errorCallback);
    },

    _getAuthHeader({url, user, pwd}) {
        let abortController = new AbortController(),
            authValue = null;

        return axios.get(url, {
            responseType: "stream",
            signal: abortController.signal
        }).then(resp => {
            abortController.abort();
            abortController.signal.onabort = () => {
                resp.data.req.abort();
                resp.data.req.destroy();
            };
            return Q.resolve(authValue);
        }, err => {
            let { status, headers } = err.response,
                authHeader = authUtils.parseWWWAuthenticateHeader(headers["www-authenticate"]),
                authMechanism = authUtils[authHeader.type.toUpperCase()];
            if (status === 401) {
                if (authMechanism) {
                    let data = {
                        username: user,
                        password: pwd
                    };
                    if (authHeader.type === "Digest") {
                        data.uri = err.request.path;
                        data.algorithm = authHeader.data.algorithm.replace("-", "").toUpperCase();
                        data.realm = authHeader.data.realm;
                        data.nonce = authHeader.data.nonce;
                        data.response = authMechanism.computeHash(data);
                    }
                    authValue = authUtils.buildAuthorizationHeader(authMechanism, data);
                } else {
                    return Q.reject(`Authentication method not supported: ${authHeader.type}`);
                }

            }
            return Q.resolve(authValue);
        });
    },

    /**
     * Stops the currently active stream
     */
    stopStream() {
        this._abortHandler && this._abortHandler.abort();
        this.imageProcessed();
    },

    /**
     * Call this when the UI is ready processing the image,
     * no further frame will be emitted to the callback until this method is called
     */
    imageProcessed() {
        this._readyForFrame = true;
    },

    /**
     * Reads the MJPEG stream and extracts every frame until VideoProxy.stopStream() is called
     * @param url
     * @param headers
     * @private
     */
    _fetchMjpgFrame({url, headers}) {
        var chunks = [],
            soi = Buffer.from([0xff, 0xd8]), // Start Of Input
            eoi = Buffer.from([0xff, 0xd9]); // End Of Input

        axios.get(url, {
            headers: headers,
            responseType: "stream",
            signal: this._abortHandler.signal
        }).then(resp => {
            resp.data.on("data", chunk => {
                if (chunk.indexOf(soi) !== -1) { // Check if the data chunk contains the JPEG start magic
                    chunks = [];
                    chunks.push(chunk.slice(chunk.indexOf(soi), chunk.length));
                } else {
                    if (chunk.indexOf(eoi) === -1) {  // The data chunk doesn't contain the JPEG end magic, add it to the received chunks
                        chunks.push(chunk);
                    } else { // The data chunk contains the JPEG end magic, extract the data until  the JPEG end magic and send the buffer to the successCallback
                        chunks.push(chunk.slice(0, chunk.indexOf(eoi) + 1 + 1)); // +1 is for the second bit of the eoi buffer, the other +1 is for get the length instead of the index
                        this._successCallback(Buffer.concat(chunks).toString("base64"));
                    }
                }
            });
            this._abortHandler.signal.onabort = () => {
                resp.data.req.abort();
                resp.data.req.destroy();
            };
        }, this._errorCallback.bind(this));
    },

    /**
     * Fetches JPG images until VideoProxy.stopStream() is called
     * @param url
     * @param headers
     * @private
     */
    _fetchJpg({url, headers}) {
        axios.get(url, {
            headers: headers,
            responseType: "arraybuffer",
            signal: this._abortHandler.signal
        }).then(resp => {
            this._fetchJpg({url, headers});
            this._successCallback(Buffer.from(resp.data).toString('base64'));
        }, this._errorCallback.bind(this));
    }
};
